function [xk,info] = PNQP(G,w,M,n,options)
%% Based on the paper:
% AN EFFICIENT QUADRATIC PROGRAMMING RELAXATION BASED ALGORITHM FOR 
% LARGE-SCALE MIMO DETECTION 
% by Ping-Fan Zhao, Qing-Na Li, Wei-Kun Chen, and Ya-Feng Liu
% to appear in SIAM Journal on Optimization, 2021.
%
% Send your comments and suggestions to    %%%%%%
%             qnl@bit.edu.cn               %%%%%%

%%%%%%%%% This version: February, 2021     %%%%%%

%% PNQP: projected Newton based quadratic penalty
% This code is designed to solve the MIMO problem, which is to restore the
% transmitted signals based on the channel matrix information and the received signals.
% The code is designed based on the relaxation of sparse quadratic programming (RSQP) model in the paper.

%% MIMO detection
% The mathematical formulation for MIMO detection is as follows.
% min ||Hx-r||_2^2 
% s.t. |x_j|^2=1, j=1,...,n;
%      arg(xj) in {0,2pi/M,...,2(M-1)pi/M}, j=1,...,n;
%
% x is an n-dimensional complex vector 

%% RSQP model
% The RSQP model for the above MIMO detection problem is as follows.
% min t^TGt+2w^Tt s.t. x>=0
% with reshape(t,M,n) column summation equals to 1
%
% t is an M*n x 1-dimensional vector


%% input:
% G is the quadratic coefficient in the objective function of the RSQP
% model ( \widetilde{G} in paper);

% w is the linear term coefficient in the objective function of the RSQP model;

% M means M-Phase-Shift Keying (M-PSK) modulation;

% n means the length of transmitted signals;

% options: parameters and other information:

% options.Mek: EM = ones(nR,nR); Mek = kron(In,EM); Mek = sparse(Mek);

% options.a and options.b: M-dimensional vectors defined as: 
% p = 0:nR-1; a = (cos(2*pi/nR.*p))'; b = (sin(2*pi/nR.*p))'; 

% options.Qr and options.Qi: Qe = H'*H; Qr = real(Qe) - diag(diag(real(Qe))); Qi = imag(Qe);
% where H is an m x n-dimensional channel matrix

%% output:
% xk: the final solution of RSQP model
% info.iter: outer iteration
% info.iter_cg: the total iteration in cg, including our own pre_cg and the matlab pcg;




display = 0; % decide whether to print iter info.
%% Try to check the res for linear system and the tolerance for identify the active set.
nM = M*n;
xk = ones(nM,1);
xk = xk./(0.2+M); % the initial point
mu = 10 ; % penalty parameter
mu_max = 1e5;
maxit = 1000;
epsilon = 0.001;     % parameter in estimating active set in projected Newton's method
eta = 0.1;           % parameter in tolerance for pcg
sigma = 1.0e-6;      % parameter in line search
beta = 0.5;          % step decreasing ration in line search
K = 1.0e8;   % upper bound to make the feasible region of subproblem upper bounded.
iter = 1;
sA0 = nM;
dk = zeros(nM,1);
Mek = options.Mek;
a = options.a;
b = options.b;
Qr = options.Qr;
Qi = options.Qi;
%% added for our own cg parameters
G = G.*2;
maxit_cg = 200;
tol_cg = 1.0e-2;
iter_cg = 0;
%preliminary calculation on the full index for future use
allIdx = (1:nM)';
allresI= mod(allIdx,M);
idx = find(allresI==0);
allresI(idx) = M;
allintI = ceil(allIdx./M);
info.iter = iter;
info.iter_cg = iter_cg;

if display
    fprintf('\n***----- PN-QP for MIMO -----***\n');
    fprintf('       n = %d, M = %d\n',n,M);
   
end

%% Algorithm starts here
[f0,GMcon,Ta,Tb,Qra,Qrb,Qia] = alfun1(xk,w,mu,M,n,Qr,Qi,a,b,0,[],[],[],[],[],[],allresI,allintI);
ourcg = 0;matcg = 0;
while iter <= maxit    
    %% Active set
    [gradLk,~,~,~,~,~,~] = alfun1(xk,w,mu,M,n,Qr,Qi,a,b,-1,Ta,Tb,Qra,Qrb,Qia,GMcon,allresI,allintI);
    omega = norm(xk - min(max(0,xk-gradLk),K));%norm
    epsk = min(epsilon,omega);
    
    ActIdx = find(xk<=epsk & gradLk(:)>=1e-10);
    InactIdx = find((xk>epsk)|(gradLk(:)<1e-10));% is faster than setdiff
    l1(iter) = length(InactIdx);

    ActIdxquit = find(xk<=0.01);
    
    sA = nM - length(ActIdxquit);
    if sA == n && sA == sA0
        if ActIdxquit0 == ActIdxquit
            break;
        end
    end
    
    ActIdxquit0 = ActIdxquit;
    sA0 = sA;
    
    %% update search direction
    xkact = xk(ActIdx);
    dkact = -xkact;
    gkact = gradLk(ActIdx);
    gkinact = gradLk(InactIdx);
    deltak = norm(gkinact);
    etak = min(eta,deltak);
    n0 =l1(iter);
    if( (n0 > 1.5*n) && nM > 1000)
        % for large scale linear system, use pcg with specially designed 
        % function Jacobian matrix to calculate Ad, for any vector d
        ourcg = ourcg+1;
        %% use our own pre-cg to save compuational cost--Qingna modified
        if (l1(iter) < nM)
            flag_fullIdx = 0;
            resI= mod(InactIdx,M);
            idx = find(resI == 0);
            resI(idx) = M;
            intI = ceil(InactIdx./M);
        else
            flag_fullIdx = 1;
            resI = allresI;
            intI = allintI;
        end
        
        [dkinact,flag1,~,tmp,res] = pre_cg(-gkinact,tol_cg,maxit_cg,ones(l1(iter),1),Qr,Qi,a,b,M,n,resI,intI,mu,deltak,flag_fullIdx);
        iter_cg = iter_cg + tmp;
        w0 = -res - dkinact*deltak - gkinact; % w0=V*d
        
    else
        matcg = matcg + 1;
        
        V = G(InactIdx, InactIdx) + Mek(InactIdx, InactIdx)*mu;
        if (l1(iter) <= 256) % for small scale linear system, use direct solver
            dkinact = (V+deltak*eye(n0))\(-gkinact);
        else % for medium scale linear system, use iteraive solver pcg in MATLAB
            [dkinact,~,~,tmp] = pcg(V + deltak*eye(n0),-gkinact,tol_cg);
            iter_cg = iter_cg + tmp;  
        end
        w0 = V*dkinact;
    end
    
    if norm(gkinact+w0) > deltak*etak && gkinact'*dkinact >= -etak*(dkinact'*dkinact)
        dkinact = -gkinact;
    end

    dk(ActIdx) = dkact;
    dk(InactIdx) = dkinact;
    %% update stepsize Armijo
    descent1 = gkinact'*dkinact;
    for stp = 1:10
        stpsz = beta^(stp - 1);
        x = max(0,xk + stpsz*dk);
        x = min(x, K);  %% upper bound to make it upper bounded.
        [fun1,GMcon,Ta,Tb,Qra,Qrb,Qia] = alfun1(x,w,mu,M,n,Qr,Qi,a,b,0,[],[],[],[],[],[],allresI,allintI);
        xtmp = min(max(0,xkact + dkact*stpsz),K);  %% make it upper bounded.
        descent2 = gkact'*(xtmp - xkact);
        dt =  stpsz*descent1 + descent2;
        if fun1-f0 <= sigma*dt
            break;
        end
    end
    xk = x;
    f0 = fun1;
    
    nmconerr = norm(GMcon);
    
    if (nmconerr > 1e-2 && mu < mu_max)
        mu = mu*3;
    end
    
    iter = iter+1;
    
end
info.iter = iter;
info.iter_cg = iter_cg;
if display
    fprintf('The algorithm stops! The final information id as follows:\n');
    fprintf('The number of iterations: %d; The number of total inner iterations in cg: %d\n', iter,iter_cg);
     fprintf('***--------------------------***\n');
end

%%  **************************************
%%  ******** All Sub-routines  ***********
%%  **************************************

%% compute function value and gradient of penalty function
    function [out,tmp_cstr,tmp_Ta,tmp_Tb,tmp_Qra,tmp_Qrb,tmp_Qia] = alfun1(x,val,mu,nr,nc,Qr,Qi,a,b,type,Ta,Tb,Qra,Qrb,Qia,cstr,allresI,allintI)
        
        switch type
            case 0
                [out,tmp_cstr,tmp_Ta,tmp_Tb,tmp_Qra,tmp_Qrb,tmp_Qia] = GMobj1(x,val,nr,nc,Qr,Qi,a,b,0,[],[],[],[],[],[],allresI,allintI);
                out = out + tmp_cstr'*tmp_cstr*mu*0.5;
            case -1
                tmp_Ta = 0;
                tmp_Tb = 0;
                tmp_Qra = 0;
                tmp_Qrb = 0;
                tmp_Qia = 0;
                
                [out,tmp_cstr,~,~,~,~,~] = GMobj1(x,val,nr,nc,Qr,Qi,a,b,-1,Ta,Tb,Qra,Qrb,Qia,cstr,allresI,allintI);
                
                tmp1 = repmat(tmp_cstr,1,nr);
                tmp1 = reshape(tmp1',nr*nc,1);
                
                out = out + tmp1.*mu;%grad
        end
    end


%% compute the function value and gradient of objective function
    function [out,tmp_cstr,tmp_Ta,tmp_Tb,tmp_Qra,tmp_Qrb,tmp_Qia] = GMobj1(x,val,nr,nc,Qr,Qi,a,b,type,Ta,Tb,Qra,Qrb,Qia,cstr,allresI,allintI)
        % Given x as vec
        switch type
            case 0
                T = reshape(x,nr,nc)';%nxM
                tmp_cstr = sum(T,2)-1;
                
                tmp_Ta = T*a;
                tmp_Tb = T*b;
                tmp_Qra = Qr*tmp_Ta;
                tmp_Qrb = Qr*tmp_Tb;
                tmp_Qia = Qi*tmp_Ta;
                out = sum(tmp_Ta.*tmp_Qra) + sum(tmp_Tb.*tmp_Qrb) + 2*sum(tmp_Tb.*tmp_Qia) + sum(val.*x)*2;% value of objective function;
            case -1
                tmp_Ta = 0;
                tmp_Tb = 0;
                tmp_Qra = 0;
                tmp_Qrb = 0;
                tmp_Qia = 0;
                tmp_cstr = cstr;
                
                h = Qra-Qi*Tb;
                o = Qia + Qrb;
                out = (kron(h,a) + kron(o,b)).*2 + val.*2; % gradient of objective function
        end
    end
end




%% PCG method
% This is exactly the algorithm by  Hestenes and Stiefel (1952)
% An iterative method to solve A(x) =b
% The symmetric positive definite matrix B is a preconditioner for A.
% See Pages 527 and 534 of Golub and va Loan (1996)
%[dkinact1,~,~,~] = pre_cg(-gkinact,maxit_cg,tol_cg,pre_c,Qr,Qi,a,b,nR,nC,InactIdx);

function [p,flag,relres,iterk,r] = pre_cg(b,tol,maxit,c,Qr,Qi,a,bb,nR,nC,resI,intI,mu,deltak,flag_fullIdx)
% Initializations
r = b;  %We take the initial guess x0=0 to save time in calculating A(x0)
n2b = norm(b);
tolb = tol * n2b; % relative tolerance
n1 = length(b);
p = zeros(n1,1);
flag = 1;
iterk = 0;
relres = 1000; % To give a big value on relres
% Precondition
z = r; %./c;  % z = B\r; here B =diag(c); if B is not the identity matrix
rz1 = r'*z;
rz2 = 1;
d = z;
% CG iteration
for k = 1:maxit
    if k > 1
        beta = rz1/rz2;
        d = z + d.*beta;
    end
    
    
    [w] = Jacobian_matrix(d,Qr,Qi,a,bb,nR,nC,resI,intI,mu,deltak,flag_fullIdx); % W =A(d)
    denom = d'*w;
    % denom = sum(d.*w);
    
    iterk = k;
    relres = norm(r)/n2b;              %relative residue = norm(r) / norm(b)
    if denom <= 0
        p = d/norm(d); % d is not a descent direction
        break % exit
    else
        alpha = rz1/denom;
        
        p = p + d.*alpha;
        r = r - w.*alpha;
        
    end
    z = r;%./c;
    if norm(r) <= tolb % Exit if Hp=b solved within the relative tolerance
        iterk = k;
        relres = norm(r)/n2b; %relative residue =norm(r) / norm(b)
        flag = 0;
        break
    end
    rz2 = rz1;
    rz1 = r'*z;
end
end

function [Vd] = Jacobian_matrix(d,Qr,Qi,a,b,nR,nC,resI,intI,mu,deltak,flag_fullIdx)
% Calculate Vd=V*d by exploring the special structure in V.
if flag_fullIdx
    matd = reshape(d,nR,nC)';
    Ad = matd*a;
    Bd = matd*b;
    cstrd = sum(matd,2);
else
    vecAd = a(resI).*d;
    matAd = sparse(intI,resI,vecAd,nC,nR);
    Ad = sum(matAd,2);
    
    vecAd = b(resI).*d;
    matAd = sparse(intI,resI,vecAd,nC,nR);
    Bd = sum(matAd,2);
    
end

Sa = Qr*Ad;
tmp = Qi*Bd;
Sa = Sa - tmp;

Sb = Qi*Ad;
tmp = Qr*Bd;
Sb = Sb + tmp;

if flag_fullIdx
    tmp11 = kron(Sa,a); tmp22 = kron(Sb,b);
    Vd = tmp11 + tmp22;
else 
    Vd = a(resI).*Sa(intI) + b(resI).*Sb(intI);
end

if flag_fullIdx
    AdI= repmat(cstrd,1,nR);
    AdI = reshape(AdI',nR*nC,1);
    
    Vd = Vd.*2;
    Vd = Vd + AdI.*mu;
else
    matAd = sparse(intI,resI, d, nC,nR);
    Ad = sum(matAd,2);
    AdI = Ad(intI);
    
    Vd = Vd.*2;
    Vd = Vd +AdI.*mu;
end
Vd = Vd + d.*deltak;
end
